from .functions import *
from ..imports import *
from .info_utils import infoRegistry

class VideoDownloader:
    def __init__(self, url, download_directory=None, user_agent=None,
                 video_extention="mp4", download_video=True,
                 get_info=False, output_filename=None,
                 ydl_opts=None, registry=None, force_refresh=False):

        self.url = get_corrected_url(url=url)
        self.registry = registry or infoRegistry()
        self.ydl_opts = ydl_opts or {}
        self.get_download = download_video
        self.get_info = get_info
        self.user_agent = user_agent
        self.video_extention = video_extention
        self.download_directory = get_download_directory(download_directory)
        self.output_filename = output_filename
        self.force_refresh = force_refresh
        self.info = {}
        self.video_urls = [self.url]

        self.send_to_dl()

    def send_to_dl(self):
        self.start()

    def build_ydl_opts(self, outtmpl):
        ydl_opts = {
            "quiet": True,
            "noprogress": True,
            "external_downloader": "ffmpeg",
            "outtmpl": outtmpl,
        }
        if self.video_extention and self.video_extention != "mp4":
            ydl_opts["format"] = (
                f"bestvideo[ext={self.video_extention}]+bestaudio[ext=m4a]/best[ext={self.video_extention}]"
            )
        else:
            ydl_opts["format"] = "bestvideo+bestaudio/best"

        if self.user_agent:
            ydl_opts["http_headers"] = {"User-Agent": self.user_agent}

        ydl_opts.update(self.ydl_opts)
        return ydl_opts

    def download(self):
        for video_url in self.video_urls:
            logger.info(f"[VideoDownloader] Processing: {video_url}")

            # 🔹 Step 1: Check registry
            cached_info = self.registry.get_video_info(video_url, force_refresh=self.force_refresh)
            if cached_info and cached_info.get("file_path") and os.path.isfile(cached_info["file_path"]):
                logger.info(f"[VideoDownloader] Found cached file: {cached_info['file_path']}")
                return cached_info

            # 🔹 Step 2: Prepare yt-dlp options
            tmp_outtmpl = os.path.join(self.download_directory, "%(id)s.%(ext)s")
            ydl_opts = self.build_ydl_opts(tmp_outtmpl)

            try:
                with yt_dlp.YoutubeDL(ydl_opts) as ydl:
                    info = ydl.extract_info(video_url, download=self.get_download)
                    video_id = info.get("id")
                    temp_path = ydl.prepare_filename(info)
                    basename = os.path.basename(temp_path)
                    _, ext = os.path.splitext(basename)

                    # 🔹 Step 3: Definitive dir layout
                    dirbase = os.path.join(self.download_directory, video_id)
                    os.makedirs(dirbase, exist_ok=True)
                    final_path = os.path.join(dirbase, f"video{ext}")

                    if temp_path != final_path:
                        shutil.move(temp_path, final_path)

                    # 🔹 Step 4: Save canonical info.json
                    info["file_path"] = final_path
                    info["video_id"] = video_id
                    info_path = os.path.join(dirbase, "info.json")
                    safe_dump_to_file(info, info_path)
                    info["info_path"] = info_path

                    # 🔹 Step 5: Update registry index
                    self.registry.edit_info(
                        data={"file_path": final_path, "info_path": info_path},
                        url=video_url,
                        video_id=video_id
                    )

                    logger.info(f"[VideoDownloader] Downloaded and stored at {final_path}")
                    self.info = info
                    return info

            except Exception as e:
                logger.error(f"[VideoDownloader] Download failed: {e}")
                return None




    def monitor(self):
        while self.monitoring:
            logger.info("Monitoring...")
            self.pause_event.wait(60)
            if self.starttime:
                elapsed_time = subtract_it(get_time_stamp(), self.starttime)
                if self.downloaded != 0 and elapsed_time != 0:
                    cumulative_time = add_it(self.downloaded, elapsed_time)
                    percent = divide_it(self.downloaded, cumulative_time)
                else:
                    percent = 0
                if percent and elapsed_time:
                    try:
                        downloaded_minutes = divide_it(elapsed_time, 60)
                        estimated_download_minutes = divide_it(downloaded_minutes, percent)
                        estimated_download_time = subtract_it(estimated_download_minutes, downloaded_minutes)
                        if estimated_download_time >= 1.5:
                            logger.info("Restarting download due to slow speed...")
                            self.start()
                    except ZeroDivisionError:
                        logger.warning("Division by zero in monitor!")
                        continue

    def start(self):
        self.download_thread = threading.Thread(target=self.download)
        self.download_thread.daemon = True
        self.monitor_thread = threading.Thread(target=self.monitor)
        self.download_thread.start()
        self.monitor_thread.start()
        self.download_thread.join()
        self.monitor_thread.join()

    def stop(self):
        self.monitoring = False
        self.pause_event.set()



from .functions import *
from ..imports import *
from abstract_utilities import get_any_value,make_list,safe_read_from_json


def get_video_info(url, ydl_opts=None,output_filename=None, cookies_path=None):
    from yt_dlp import YoutubeDL

    
    ydl_opts = {
        'quiet': True,
        'skip_download': True,
    }

    if cookies_path and os.path.exists(cookies_path):
        ydl_opts['cookiefile'] = cookies_path

    try:
        with YoutubeDL(ydl_opts) as ydl:
            info = ydl.extract_info(url, download=False)
            return info
    except Exception as e:
        print(f"Failed to extract video info: {e}")
        return None
class infoRegistry(metaclass=SingletonMeta):
    """Singleton for managing video + info registry directories."""

    def __init__(self, video_directory=None, envPath=None, info_directory=None,temp_directory=None,data_directory=None,**kwargs):
        if not hasattr(self, 'initialized'):
            self.initialized = True
            self.videoDirectory = get_video_directory(video_directory=video_directory, envPath=envPath)
            self.infoRegistryDirectory = get_info_directory(directory=info_directory, envPath=envPath)
            self.tempDirectory = get_info_directory(directory=temp_directory, envPath=envPath)
            self.dataDirectory = get_info_directory(directory=data_directory, envPath=envPath)
            self.registry_path = os.path.join(self.infoRegistryDirectory, "registry.json")
            self._load_registry()

    def _load_registry(self):
        if os.path.isfile(self.registry_path):
            with open(self.registry_path, "r", encoding="utf-8") as f:
                self.registry = json.load(f)
        else:
            self.registry = {"by_url": {}, "by_id": {}}

    def _save_registry(self):
        with open(self.registry_path, "w", encoding="utf-8") as f:
            json.dump(self.registry, f, indent=2)

    def _get_cache_path(self, video_id=None,url=None):
        video_id = self._get_video_id(url=url,video_id=video_id)
        return os.path.join(self.infoRegistryDirectory, f"{video_id}.json")

    def _get_video_id(self,url=None,video_id=None):
        if url and not video_id:
            video_id = self.registry["by_url"].get(url)
        return video_id
    def _fetch_get_id(self,url=None):
        ydl_opts = {"quiet": True, "skip_download": True}
        info = get_video_info(url,ydl_opts=ydl_opts)
        video_id = info.get("id") or hashlib.sha1(url.encode("utf-8")).hexdigest()[:12]
        cache_path = self._get_cache_path(video_id)
        safe_dump_to_file(data=info,file_path=cache_path)
        return video_id,cache_path
    def _save_to_info_path(self,data,url=None,video_id=None):
        if not video_id and not url:
            video_id = data.get("id")
        if not video_id and url:
            video_id = data.get("id") or hashlib.sha1(url.encode("utf-8")).hexdigest()[:12]
        if video_id:
            cache_path = self._get_cache_path(video_id)
            safe_dump_to_file(data=data,file_path=cache_path)
            return video_id,cache_path
    def _get_cached_info(self,url=None,video_id=None):
        cache_path = self._get_cache_path(url=url,video_id=video_id)
        if os.path.isfile(cache_path):
            with open(cache_path, "r", encoding="utf-8") as f:
                return json.load(f)
    def _add_url_top_registry(self,url,video_id):
        if url:
            self.registry["by_url"][url] = video_id
        self.registry["by_id"][video_id] = {"url": url, "timestamp": time.time()}
        self._save_registry()
    def get_video_info(self, url=None, video_id=None, force_refresh=False,download=False):
        url = get_corrected_url(url=url)
        if not (url or video_id):
            raise ValueError("Either url or video_id must be provided")

        # 1. Lookup video_id from registry if only URL given
        video_id = video_id or self._get_video_id(url=url)

        # 2. If video_id known, check cache
        if video_id:
            cache_path = self._get_cache_path(video_id)
            if os.path.isfile(cache_path) and not force_refresh:
                return self._get_cached_info(url=url,video_id=video_id)
                
        # 3. Otherwise, fetch with yt_dlp
        video_id,cache_path = self._fetch_get_id(url=url)
        # Update registry
        self._add_url_top_registry(url=url,video_id=video_id)

        return self._get_cached_info(url=url,video_id=video_id)
    def edit_info(self,data,url=None,video_id=None, force_refresh=False,download=False):
        video_info = self.get_video_info(url=url, video_id=video_id, force_refresh=force_refresh,download=download)
        video_info.update(data)
        self._save_to_info_path(data=video_info,url=url,video_id=video_id)
        return video_info
    def list_cached_videos(self):
        """Return all cached entries (url, id, metadata)."""
        results = []
        for vid, meta in self.registry["by_id"].items():
            results.append({"video_id": vid, "url": meta["url"], "timestamp": meta["timestamp"]})
        return results



